import { Direction, HandleKeyEventOptions, InsertTreeOptions, KeyEvent, MoveNodeOptions, Node, NodeConfig, NodeId, NodesBag, RegisterOverrideOptions, UnregisterNodeOptions } from './interfaces';
import mitt from 'mitt';
export * from './interfaces';
export declare class Lrud {
    nodes: NodesBag;
    rootNode: Node;
    currentFocusNode?: Node;
    isIndexAlignMode: boolean;
    emitter: mitt.Emitter;
    constructor();
    /**
     * Registers a callback for an LRUD event.
     *
     * @param {string} eventName - event to subscribe to
     * @param {function} callback - function to call on event
     */
    on(eventName: string, callback: mitt.Handler): void;
    /**
     * Unregisters a callback for an LRUD event.
     *
     * @param {string} eventName - event to unsubscribe from
     * @param {function} callback - function that was added using .on()
     */
    off(eventName: string, callback: mitt.Handler): void;
    /**
     * Returns the root node.
     */
    getRootNode(): Node;
    /**
     * Returns the current focused node.
     */
    getCurrentFocusNode(): Node | undefined;
    /**
     * Registers a new node into the LRUD tree.
     *
     * @param {string} nodeId - id of the node to register
     * @param {object} [nodeConfig] - registered node parameters
     * @param {string} [nodeConfig.parent] - parent node id, if null, default root node is used
     * @param {number} [nodeConfig.index] - if null, index is 1 more than the index of the last sibling. if no previous siblings, index is 1
     * @param {number[]} [nodeConfig.indexRange] - defaults to null. acts as a colspan, value [0] is lower bound, value [1] is upper bound
     * @param {object} [nodeConfig.selectAction] - if a node has a selectAction, it is focusable
     * @param {boolean} [nodeConfig.isFocusable] - if a node is explicitly set as isFocusable, it is focusable
     * @param {boolean} [nodeConfig.isWrapping] - if true, when asking for the next child at the end or start of the node, the will "wrap around" and return the first/last (when asking for the last/first)
     * @param {string} [nodeConfig.orientation] - can be "vertical" or "horizontal". is used in conjunction when handling direction of key press, to determine which child is "next"
     * @param {boolean} [nodeConfig.isIndexAlign]  -if a node is index aligned, its descendents should jump to nodes based on index instead of activeChild
     * @param {function} [nodeConfig.onLeave] - if a node has an `onLeave` function, it will be run when a move event leaves this node
     * @param {function} [nodeConfig.onEnter] - if a node has an `onEnter` function, it will be run when a move event enters this node
     */
    registerNode(nodeId: NodeId, nodeConfig?: NodeConfig): Lrud;
    /**
     * Allows to change node's parent by moving it's whole sub-tree.
     *
     * @param {string|object} node - node or id of the node that is about to change parent
     * @param {string|object} newParentNode - node or id of the node that became a new parent for nodeId
     * @param {object} [options]
     * @param {number} [options.index] - index at which nodeId should be inserted as a child of newParentNodeId
     * @param {boolean} [options.maintainIndex] - applies only if index is not defined; if true, node will be inserted at
     *                                           it's current position if possible; otherwise node will be appended; default: false
     */
    moveNode(node: NodeId | Node, newParentNode: NodeId | Node, options?: MoveNodeOptions): void;
    /**
     * Registers a new node into the LRUD tree.
     *
     * Kept for backwards compatibility reasons.
     *
     * @param {string} nodeId - id of the node to register
     * @param {object} [nodeConfig] - registered node parameters
     */
    register(nodeId: NodeId, nodeConfig?: NodeConfig): Lrud;
    /**
     * Unregisters a node from the navigation tree.
     *
     * Kept for backwards compatibility reasons.
     *
     * @param {string|object} node - node or id of the node to unregister
     * @param {object} [unregisterOptions]
     */
    unregister(node: NodeId | Node, unregisterOptions?: UnregisterNodeOptions): void;
    /**
     * Unregisters a node from the navigation tree.
     *
     * @param {string|object} node - node or id of the node to unregister
     * @param {object} [unregisterOptions]
     * @param {boolean} [unregisterOptions.forceRefocus] if true, LRUD will attempt to re-focus on a new node if the currently focused
     *                                                   node becomes unregistered due to the given node ID being unregistered
     */
    unregisterNode(node: NodeId | Node, unregisterOptions?: UnregisterNodeOptions): Lrud;
    /**
     * Registers a new override onto the LRUD instance.
     *
     * @param {string|object} source - node or id of the node for which override should be triggered
     * @param {string|object} target - node or id of the node to which this overrides points
     * @param {string} direction - traversal direction, for which this override should be triggered
     * @param {object} [options]
     * @param {boolean} [options.forceOverride] if true, existing override from source node in direction will be overwritten.
     */
    registerOverride(source: NodeId | Node, target: NodeId | Node, direction: Direction, options?: RegisterOverrideOptions): Lrud;
    /**
     * Unregisters an override from the LRUD instance.
     *
     * @param {string|object} source - node or id of the node for which override should be unregistered
     * @param {string} [direction] - traversal direction, in which override should be unregistered
     */
    unregisterOverride(source: NodeId | Node, direction?: Direction): Lrud;
    /**
     * Returns a node for an ID.
     *
     * @param {string} nodeId - id of the node to be returned
     */
    getNode(nodeId: NodeId): Node | undefined;
    /**
     * Gets a node by ID and then unregisters it from the instance.
     *
     * @param {string} nodeId - id of the node to be picked
     */
    pickNode(nodeId: NodeId): Node | undefined;
    /**
     * Starting from a node, climb up the navigation tree until we find a node that can be
     * actioned, based on the given direction. An actionable node is one whose orientation is valid
     * for the given direction, has focusable children and whose activeChild isn't a leaf that is
     * also its current activeChild.
     *
     * @param {object} node - node from which climbing up starts
     * @param {string} direction - direction in which to traverse while climbing up
     */
    climbUp(node: Node, direction: Direction): Node | undefined;
    /**
     * Starting from the given node, dig down the navigation tree until we find a focusable
     * leaf, and return it. dig "direction" priority:
     * - index align mode
     * - active child
     * - first focusable child
     *
     * @param {object} node - node, from which digging down starts
     * @param {string} [direction] - direction in which to traverse while digging down
     */
    digDown(node: Node, direction?: Direction): Node | undefined;
    /**
     * Gets the semantic next focusable child for a given direction.
     *
     * If the direction is 'left' or 'up', return the semantic previous focusable child of the node.
     * If the direction is 'right' or 'down'', return the semantic next focusable child of the node.
     * If the direction is *, return the semantic next (or previous, if next not found) focusable child of the node.
     *
     * @param {object} node - node, for which next focusable child for a given direction is returned
     * @param {string} direction - direction in which to traverse while searching for next focusable child
     */
    getNextFocusableChildInDirection(node: Node, direction: Direction): Node | undefined;
    /**
     * Gets the semantic "next" child for a node that might be focused or bypass focus to its children.
     *
     * @param {object} node - node, for which next focusable child is returned
     */
    getNextFocusableChild(node: Node): Node | undefined;
    /**
     * Gets the semantic "previous" child for a node that might be focused or bypass focus to its children.
     *
     * @param {object} node - node, for which previous focusable child is returned
     */
    getPrevFocusableChild(node: Node): Node | undefined;
    /**
     * Gets the first child of a node, based on index.
     *
     * @param {object} node - node, for which first child is returned
     */
    getNodeFirstChild(node: Node): Node | undefined;
    /**
     * Gets the last child of a node, based on index.
     *
     * @param {object} node - node, for which last child is returned
     */
    getNodeLastChild(node: Node): Node | undefined;
    /**
     * Gets the first focusable (or containing focusable nodes) child of a node, based on index.
     *
     * @param {object} node - node, for which first focusable child is returned
     */
    getNodeFirstFocusableChild(node: Node): Node | undefined;
    /**
     * Gets the last focusable (or containing focusable nodes) child of a node, based on index.
     *
     * @param {object} node - node, for which last focusable child is returned
     */
    getNodeLastFocusableChild(node: Node): Node | undefined;
    /**
     * Given an event, handle any state changes that may arise from the direction pressed.
     *
     * State changes based on climbing up and digging down from the current focusedNode
     *
     * @param {object} event
     * @param {number} [event.keyCode]
     * @param {string} [event.direction]
     * @param {object} [options]
     * @param {boolean} [options.forceFocus] - if true and there's no currently focused node, LRUD will try to focus
     *                                         first focusable node; default: false
     */
    handleKeyEvent(event: KeyEvent, options?: HandleKeyEventOptions): Node | undefined;
    /**
     * Sets the activeChild of the parent node to the value of the child node.
     *
     * @param {string|object} parent - node or id of the node, which activeChild is about to be set
     * @param {string|object} child - node or id of the node, that is about to be set as parent's activeChild
     */
    setActiveChild(parent: NodeId | Node, child: NodeId | Node): void;
    /**
     * Sets the activeChild of the parent node to the value of the child node.
     *
     * If the parent node has a parent itself, it digs up the tree and sets those activeChild values.
     *
     * @param {string|object} parent - node or id of the node, which activeChild is about to be set
     * @param {string|object} child - node or id of the node, that is about to be set as parent's activeChild
     */
    setActiveChildRecursive(parent: NodeId | Node, child: NodeId | Node): void;
    /**
     * Unsets the activeChild of the parent nodes branch ensuring that activeChild is on the unsetting child node path
     * and not on the currentFocusNode's path, unless child node is a currentFocusNode.
     *
     * @param {string|object} parent - node or id of the node, which activeChild is about to be unset
     * @param {string|object} activeChild - node or id of the node, that is about to be unset as parent's activeChild
     */
    unsetActiveChild(parent: NodeId | Node, activeChild: NodeId | Node): void;
    /**
     * Sets the current focus of the instance to the given node or node ID.
     *
     * If the given node points to a non-focusable node, we dig down from
     * the given node to find a node that can be focused on.
     *
     * Calls `onFocus` on the given node, if it exists, and emits a `focus` event,
     * also calls `onBlur` on the node that WAS focused before this function was called.
     *
     * @param {string|object} node - node or id of the node to be focused
     */
    assignFocus(node: NodeId | Node): void;
    /**
     * If the focus of the tree is out of sync, ie, the current focused node becomes unfocusable,
     * this can be used to fall back to another focus.
     *
     * @param {object} node - node, based on which focus is recalculated
     */
    recalculateFocus(node: Node): void;
    /**
     * Given a tree, register all of its nodes into this instance.
     *
     * @param {object} subTreeRootNodeConfig
     */
    registerTree(subTreeRootNodeConfig: NodeConfig): void;
    /**
     * Given a tree object, attempt to register that tree into the current lrud instance.
     *
     * If the given tree already exists as a branch in the instance tree, the new tree will replace that branch.
     *
     * If the new tree doesn't already exist as a branch in the instance tree, this function will register the new
     * tree as a branch against the root node, as per standard registerNode() behaviour.
     *
     * @param {object} subTreeRootNodeConfig
     * @param {object} [options]
     * @param {boolean} [options.maintainIndex] - if true, and new tree is replacing an existing branch of the tree,
     *                                            maintain the original branches relative index; default: true
     */
    insertTree(subTreeRootNodeConfig: NodeConfig, options?: InsertTreeOptions): void;
    /**
     * Checks is node contains children that might be focused (are a focusable candidates).
     * It checks the whole node's children sub-tree, not only direct children.
     *
     * @param {object} node - node, which children are checked against being focusable candidates
     */
    doesNodeHaveFocusableChildren(node: Node): boolean;
    /**
     * Checks if node is focusable or contains focusable children.
     *
     * @param {object} node - node to check against being focusable candidate
     */
    isNodeFocusableCandidate(node: Node): boolean;
    /**
     * Checks if given parent node is a child's parent node or the node itself. If it's a parent it doesn't have
     * to be a direct one, but has to be placed at a path to the root node.
     *
     * E.g.
     *        root
     *        / \
     *       A   B
     *      /     \
     *     AA      BA
     *    /  \
     *  AAA  AAB
     *
     *  Expect:
     *    isSameOrParentForChild('A', 'A') -> true
     *    isSameOrParentForChild('A', 'AA') -> true
     *    isSameOrParentForChild('AA', 'A') -> false
     *    isSameOrParentForChild('A', 'AAA') -> true
     *    isSameOrParentForChild('AAA', 'AA') -> false
     *    isSameOrParentForChild('A', 'BA') -> false
     *    isSameOrParentForChild('BA', 'A') -> false
     *    isSameOrParentForChild('AA', 'BA') -> false
     *    isSameOrParentForChild('BA', 'AA') -> false
     *
     * @param {string|object} parent - parent or id of the examined parent
     * @param {string|object} child - node id of the node which parents are queried
     */
    isSameOrParentForChild(parent: NodeId | Node, child: NodeId | Node): boolean;
    /**
     * Changes the ability of a node to be focused in place.
     *
     * @param {string|object} node - node or id of the node, which focusability property is about to be changed
     * @param {boolean} isFocusable - focusability value to set
     */
    setNodeFocusable(node: NodeId | Node, isFocusable: boolean): void;
}
