import { Direction, Node, NodeConfig, NodeId, NodeIndex, NodeIndexRange, Orientation, Tree } from './interfaces';
/**
 * Given an array of values and a goal, return the value from values which is closest to the goal.
 *
 * @param {number[]} values
 * @param {number} goal
 */
export declare const closestIndex: (values: NodeIndex[], goal: NodeIndex) => number;
/**
 * Checks if a given node is focusable.
 *
 * @param {object} node - node to check against focusability
 */
export declare const isNodeFocusable: (node: Node) => boolean;
/**
 * Given a keyCode, lookup and return the direction from the keycodes mapping file.
 *
 * @param {number} keyCode - key code for which corresponding direction is searched
 */
export declare const getDirectionForKeyCode: (keyCode: number) => Direction | undefined;
/**
 * Given an orientation and a direction, do they match?
 *
 * I.e an orientation `horizontal` and direction `left` or `right` is considered matching.
 *
 * Direction CAN be passed as `*` (wildcard). this will always return true.
 *
 * @param {string} orientation - orientation to match with the direction
 * @param {string} direction - direction to match with the orientation
 */
export declare const isDirectionAndOrientationMatching: (orientation: Orientation, direction: Direction) => boolean;
/**
 * Returns a child from the given node whose indexRange encompasses the given index.
 *
 * @param {object} node - node, which children index ranges will be examined
 * @param {number} index - examined index value
 */
export declare const findChildWithMatchingIndexRange: (node: Node, index: NodeIndex) => Node | undefined;
/**
 * Returns a child from the given node whose index is numerically closest to the given index.
 *
 * If an indexRange is provided, first check if the node's activeChild is inside
 * the indexRange. If it is, return the activeChild node instead
 *
 * @param {object} node
 * @param {number} index
 * @param {number[]} indexRange
 */
export declare const findChildWithClosestIndex: (node: Node, index: NodeIndex, indexRange?: NodeIndexRange) => Node | undefined;
/**
 * Inserts given child to the parent's children, keeping indices coherent and compact.
 *
 * @param parent - node to which new child is about to be added
 * @param newChild - node to be added to the parent
 */
export declare const insertChildNode: (parent: Node, newChild: Node) => void;
/**
 * Removes given child from the parent's children, keeping indices coherent and compact.
 *
 * @param parent - node from which children given child is about to be removed
 * @param child - node to be removed from parent's children
 */
export declare const removeChildNode: (parent: Node, child: Node) => void;
/**
 * Returns valid orientation parameter.
 *
 * TypeScript defines type appropriately, but in JavaScript anything can be passed. Method doesn't imply that user
 * might be malicious, it just tries to unify provided string values, by making them a valid Orientation type.
 *
 * @param orientation - orientation to correct
 */
export declare const toValidOrientation: (orientation: Orientation) => Orientation | undefined;
/**
 * Returns valid direction parameter.
 *
 * TypeScript defines type appropriately, but in JavaScript anything can be passed. Method doesn't imply that user
 * might be malicious, it just tries to unify provided string values, by making them a valid Orientation type.
 *
 * @param direction - direction to correct
 */
export declare const toValidDirection: (direction: Direction) => Direction | undefined;
/**
 * Creates node object from given node parameters.
 *
 * Node creation method is optimized for JavaScript engine.
 * Objects created in "the same way" allows JavaScript engine reusing existing HiddenClass transition chain.
 * Moreover most used properties are defined "at the beginning" making access to them a bit faster.
 *
 * @param {string} nodeId - id of the node
 * @param {object} [nodeConfig] - node parameters
 */
export declare const prepareNode: (nodeId: NodeId, nodeConfig?: NodeConfig) => Node;
/**
 * Traverses through node subtree (including the node) with iterative preorder deep first search tree traversal algorithm.
 *
 * DFS is implemented without recursion to avoid putting methods to the stack. This allows traversing deep trees without
 * the need of allocating memory for recursive method calls.
 *
 * For some processes, when node meeting some condition is searched, there's no need to traverse through whole tree.
 * To address that, given nodeProcessor may return boolean value. when true is returned, than traversal algorithm
 * is interrupted immediately.
 *
 * E.g. Given tree:
 *        root
 *        / \
 *       A   B
 *      /     \
 *     AA      BA
 *    /  \
 *  AAA  AAB
 *
 *  Traversal path: root -> A -> AA -> AAA -> AAB -> B -> BA
 *
 * @param {object} startNode - node that is the root of traversed subtree
 * @param {function} nodeProcessor - callback executed for traversed node, if true is returned then subtree traversal is interrupted
 */
export declare const traverseNodeSubtree: <NodeType extends Tree<NodeType>>(startNode: NodeType, nodeProcessor: (node: NodeType) => boolean | void) => void;
